import os
import re

from boil.utils.file_utils import ensure_dir


class TemplateRenderer:

    PATH_EXCLUDE_PATTERNS = [r"^_includes\/.+"]

    def __init__(self, env, vars, target_dir):
        self._env = env
        self._vars = vars
        self._target_dir = target_dir

    def render(self):
        for template_path in self._env.list_templates():
            if self._is_valid_path(template_path):
                self._render_template(template_path)

    def _is_valid_path(self, template_path):
        for exclude_pattern in self.PATH_EXCLUDE_PATTERNS:
            if re.match(exclude_pattern, template_path):
                return False
        return True

    def _render_template(self, template_path):
        template = self._env.get_template(template_path)

        target_path = self._build_target_path(template_path)
        ensure_dir(target_path)

        with open(target_path, "wb") as target:
            target.write(template.render(self._vars).encode("utf8"))

    def _build_target_path(self, template_path):
        target_path = os.path.join(self._target_dir, template_path)

        target_path_template = self._env.from_string(target_path)
        target_path = target_path_template.render(self._vars)

        if target_path.endswith(".j2"):
            target_path = os.path.splitext(target_path)[0]
        return target_path



from boil.common.filters import humanize
from boil.utils.display import display, prompt


class VariableLoader:
    def __init__(self, vars):
        self._vars = vars

    def get_vars(self):
        collected_vars = {}
        for var in self._vars:
            collected_vars[var.name] = self._collect_var(var)
        return collected_vars

    def _collect_var(self, var):
        prompt_msg = self._build_prompt_msg(var)
        value = prompt(prompt_msg)
        if not value:
            default = var.default
            if default:
                display("Using default: %s." % default)
                value = default
        while var.required and not value:
            display("Value required.", color="red")
            value = prompt(prompt_msg)
        return value

    def _build_prompt_msg(self, var):
        var_name = humanize(var.name)

        example = var.example
        default = var.default

        hint = None
        if default:
            hint = default
        elif example:
            hint = "e.g. %s" % example

        prompt_msg = None
        if hint:
            prompt_msg = "%s [%s]:\n" % (var_name, hint)
        else:
            prompt_msg = "%s:\n" % var_name
        return prompt_msg



import pkg_resources

from boil.exceptions import PlateNotFound
from boil.plate import Plate


class PlateManager:

    ENTRY_NAMESPACE = "boil.plates"

    def __init__(self):
        self._plates = self._load_plates()

    def get_plate(self, plate_name):
        if plate_name not in self._plates:
            raise PlateNotFound(name=plate_name)
        return self._plates[plate_name]

    def get_plate_names(self):
        return [plate.name for plate in self._plates.values()]

    def _load_plates(self):
        plates = {}
        for entry_point in self._iter_entry_points():
            plate_module = entry_point.load()
            plates[entry_point.name] = Plate(plate_module)
        return plates

    def _iter_entry_points(self):
        return pkg_resources.iter_entry_points(self.ENTRY_NAMESPACE)



import jinja2

from boil.common import filters


class Environment(jinja2.Environment):
    def __init__(self, package_name):
        loader = jinja2.PackageLoader(package_name)
        super(Environment, self).__init__(
            loader=loader,
            keep_trailing_newline=True,
            trim_blocks=True,
            lstrip_blocks=True,
        )
        self._setup_filters()

    def _setup_filters(self):
        self.filters.update(filters.TEMPLATE_FILTERS.copy())



import os
import re

from boil.common.filters import humanize
from boil.plate.manager import PlateManager
from boil.plate.runner import PlateRunner
from boil.utils.display import display, display_list


class Command:
    def execute(self, args):
        raise NotImplementedError()


class ListPlates(Command):
    def execute(self, args):
        plate_names = self._list_plates()
        display("Available plates:", bold=True)
        display_list(sorted(plate_names))

    def _list_plates(self):
        return PlateManager().get_plate_names()


class SearchPlates(Command):
    def execute(self, args):
        phrase = args["<phrase>"]
        plate_names = self._search_plates(phrase)
        num_found = len(plate_names)
        if num_found > 0:
            display("Found plates (%s):" % num_found, bold=True)
            display_list(sorted(plate_names))
        else:
            display("No plates found.")

    def _search_plates(self, phrase):
        plate_names = PlateManager().get_plate_names()
        return [name for name in plate_names if phrase in name]


class RunPlate(Command):
    def execute(self, args):
        plate_name = self._normalize_name(args["<plate_name>"])
        target_dir = args.get("--target-dir") or os.getcwd()

        display("Initializing new %s." % humanize(plate_name))
        self._run_plate(plate_name, target_dir)
        display("Done!", color="green")

    def _normalize_name(self, name):
        return re.sub(r"\W", "_", name)

    def _run_plate(self, plate_name, target_dir):
        plate = PlateManager().get_plate(plate_name)
        PlateRunner(plate, target_dir).run()


def get_command(args):
    cmd_class = get_command_class(args)
    return cmd_class()


def get_command_class(args):
    if args["list"]:
        return ListPlates
    elif args["search"]:
        return SearchPlates
    elif args["new"]:
        return RunPlate



"""Boil

Usage:
    boil list
    boil search <phrase>
    boil new <plate_name> [--target-dir=<dir>]
    boil -h | --help

Options:
    --target-dir=<dir>  Target directory where project files should be
                        populated.
"""

from docopt import docopt

from boil.cli import commands as cmd
from boil.exceptions import BoilError
from boil.utils.display import display


def main():
    args = docopt(__doc__)

    try:
        cmd.get_command(args).execute(args)
    except BoilError as ex:
        display(
            "An error ocurred while executing the command: %s" % str(ex), color="red"
        )



import math
import sys

from six.moves import input

from boil.utils import ansi


def display(msg, color=None, bold=False, stderr=False):
    msg = ansi.format(msg, color=color, bold=bold)
    msg += "\n"

    file_obj = sys.stdout
    if stderr:
        file_obj = sys.stderr

    file_obj.write(msg)
    file_obj.flush()


def display_list(obj, cols=4, columnwise=True, gap=10):
    slist = [str(item) for item in obj]
    num_items = len(slist)

    cols = min(cols, num_items)
    max_len = max([len(item) for item in slist])

    if columnwise:
        cols = int(math.ceil(float(num_items) / cols))

    plist = [slist[i : i + cols] for i in range(0, num_items, cols)]

    if columnwise:
        plist[-1].extend([""] * (num_items - len(plist[-1])))
        plist = zip(*plist)

    segments = ["".join([c.ljust(max_len + gap) for c in p]) for p in plist]
    msg = "\n".join(segments)
    display(msg)


def prompt(msg=None):
    if msg:
        msg = ansi.format(msg, bold=True)
    return input(msg)



class Variable:
    def __init__(self, name, required=False, default=None, example=None):
        self.name = name
        self.required = required
        self.default = default
        self.example = example


def parse_vars(var_defs):
    return [Variable(**var_def) for var_def in var_defs]



import os

from boil.exceptions import ArtifactAlreadyExists
from boil.template.renderer import TemplateRenderer
from boil.template.environment import Environment
from boil.utils.file_utils import copy_tree, temp_dir
from boil.vars.loader import VariableLoader


class PlateRunner:

    """
    This is the primary class responsible for running plates. It generates
    project file structure inside a temporary directory and copies its contents
    to the destination directory afterwards, ensuring that there are no
    conflicts with already existing artifacts.
    """

    def __init__(self, plate, target_dir, overwrite=False):
        self._plate = plate
        self._target_dir = target_dir
        self._overwrite = overwrite

    def run(self):
        vars = VariableLoader(self._plate.vars).get_vars()
        env = Environment(self._plate.module_name)

        with temp_dir() as tmp_dir:
            TemplateRenderer(env, vars, tmp_dir).render()

            if not self._overwrite:
                target_names = os.listdir(self._target_dir)
                for artifact_name in os.listdir(tmp_dir):
                    if artifact_name in target_names:
                        raise ArtifactAlreadyExists(name=artifact_name)

            copy_tree(tmp_dir, self._target_dir)



ANSI_CODES = {
    "black": "30",
    "white": "37",
    "blue": "34",
    "green": "32",
    "cyan": "36",
    "red": "31",
    "purple": "35",
    "yellow": "33",
    "magenta": "35",
    "bold": "1",
    "reset": "0",
}

ANSI_CODE_PATTERN = "\33[%sm"


def format(text, color=None, bold=False):
    styles = []
    if color:
        styles.append(color)
    if bold:
        styles.append("bold")

    start_code = "".join([ansi_code(style) for style in styles])
    end_code = ansi_code("reset")

    formatted = []
    for line in text.split("\n"):
        formatted.append("%s%s%s" % (start_code, line, end_code))

    return "\n".join(formatted)


def ansi_code(style):
    return ANSI_CODE_PATTERN % ANSI_CODES[style]



VARS = [
    {"name": "license_type", "required": True, "default": "mit"},
    {"name": "product_name", "required": True, "example": "Super fancy product"},
    {"name": "author", "required": True, "example": "John Smith or name of a company"},
    {"name": "year", "required": True, "example": "2018"},
]



import inflection


camelize = inflection.camelize
underscore = inflection.underscore
dasherize = inflection.dasherize
humanize = inflection.humanize
titleize = inflection.titleize
parameterize = inflection.parameterize
pluralize = inflection.pluralize
singularize = inflection.singularize
ordinal = inflection.ordinal
ordinalize = inflection.ordinalize


TEMPLATE_FILTERS = {
    "camelize": camelize,
    "underscore": underscore,
    "dasherize": dasherize,
    "humanize": humanize,
    "titleize": titleize,
    "parameterize": parameterize,
    "pluralize": pluralize,
    "singularize": singularize,
    "ordinal": ordinal,
    "ordinalize": ordinalize,
}



class BoilError(Exception):

    message = "An unknown exception occurred."

    def __init__(self, **kwargs):
        msg = self.message % kwargs
        super(BoilError, self).__init__(msg)


class PlateNotFound(BoilError):

    message = "Plate '%(name)s' not found."


class ArtifactAlreadyExists(BoilError):

    message = "Artifact '%(name)s' already exists in target directory."



import contextlib
import os
import shutil
import tempfile


def ensure_dir(path):
    dir = os.path.dirname(path)
    if not os.path.exists(dir):
        os.makedirs(dir)


@contextlib.contextmanager
def temp_dir():
    tmp_dir = tempfile.mkdtemp()
    try:
        yield tmp_dir
    finally:
        shutil.rmtree(tmp_dir)


def copy_tree(src, dst):
    for name in os.listdir(src):
        src_name = os.path.join(src, name)
        dst_name = os.path.join(dst, name)
        if os.path.isdir(src_name):
            shutil.copytree(src_name, dst_name)
        else:
            shutil.copy2(src_name, dst_name)



VARS = [
    {"name": "package_name", "required": True, "example": "fancy_package"},
    {"name": "description", "example": "Super fancy package"},
    {"name": "url", "example": "https://github.com/username/package_name"},
    {"name": "license", "default": "MIT"},
    {"name": "author", "example": "John Smith"},
]



from boil.vars import parse_vars


class Plate:
    def __init__(self, module):
        self.module = module
        self._load_data()

    def _load_data(self):
        module_name = self.module.__name__

        self.name = module_name.split(".")[-1]
        self.module_name = module_name

        self.vars = []
        self.filters = {}

        if hasattr(self.module, "VARS"):
            self.vars = parse_vars(self.module.VARS)

        if hasattr(self.module, "FILTERS"):
            self.filters = self.module.FILTERS



VARS = [
    {"name": "plate_name", "required": True, "example": "fancy_plate"},
    {"name": "description", "required": True, "example": "Super fancy plate"},
    {"name": "author", "example": "John Smith"},
]



VARS = [
    {"name": "role_name", "required": True, "example": "redis"},
    {"name": "description", "example": "Role for setting Redis instance up"},
    {"name": "license", "default": "MIT"},
    {"name": "author", "example": "John Smith"},
]



import os
import sys
from shutil import rmtree

from setuptools import setup, find_packages, Command


base_dir = os.path.abspath(os.path.dirname(__file__))

pkg_metadata = {}
with open(os.path.join(base_dir, "boil", "__version__.py")) as f:
    exec(f.read(), pkg_metadata)


def get_description():
    with open("README.md") as readme:
        return readme.read()


def get_requirements():
    with open("requirements.txt") as requirements:
        return requirements.read().splitlines()


class UploadCommand(Command):

    """Extends setup.py with a command to build and publish the package."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds...")
            rmtree(os.path.join(base_dir, "dist"))
        except OSError:
            pass
        self.status("Building Source distribution...")
        os.system("{0} setup.py sdist bdist_wheel".format(sys.executable))
        self.status("Uploading the package to PyPi via Twine...")
        os.system("twine upload dist/*")
        self.status("Pushing git tags...")
        os.system("git tag v{0}".format(pkg_metadata["__version__"]))
        os.system("git push --tags")
        sys.exit()


setup(
    name="boil",
    version=pkg_metadata["__version__"],
    description="Radically simple app initialization",
    long_description_content_type="text/markdown",
    long_description=get_description(),
    url="https://github.com/bzurkowski/boil",
    author="Bartosz Zurkowski",
    license="MIT",
    install_requires=get_requirements(),
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "boil = boil.cli:main",
        ],
        "boil.plates": [
            "python_package = boil.plates.python_package",
            "ansible_role = boil.plates.ansible_role",
            "bash_script = boil.plates.bash_script",
            "license = boil.plates.license",
            "ruby_gem = boil.plates.ruby_gem",
            "plate = boil.plates.plate",
        ],
    },
    classifiers=[
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3.9",
        "Topic :: System :: Installation/Setup",
        "Topic :: System :: Systems Administration",
        "Topic :: Utilities",
    ],
    include_package_data=True,
    zip_safe=False,
    cmdclass={"upload": UploadCommand},
)



VARS = [
    {"name": "gem_name", "required": True, "example": "fancy_gem"},
    {"name": "ruby_version", "required": True, "example": "2.5.1"},
    {"name": "description", "required": True, "example": "Super fancy gem"},
    {"name": "author", "required": True, "example": "John Smith"},
    {"name": "url", "example": "https://github.com/username/gem_name"},
    {"name": "license", "default": "MIT"},
    {"name": "email"},
]
